import numpy as numpy

from .Kernel import Kernel

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  *    This file is part of the BayesicFitting package.
#  *
#  *    BayesicFitting is free software: you can redistribute it and/or modify
#  *    it under the terms of the GNU Lesser General Public License as
#  *    published by the Free Software Foundation, either version 3 of
#  *    the License, or ( at your option ) any later version.
#  *
#  *    BayesicFitting is distributed in the hope that it will be useful,
#  *    but WITHOUT ANY WARRANTY; without even the implied warranty of
#  *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  *    GNU Lesser General Public License for more details.
#  *
#  *    The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2017 - 2018 Do Kester

class Tophat( Kernel ):
    """
    Tophat is a Kernel function which is 1.0 between [-0.5,0.5]; it is 0 elsewhere.

    Attributes
    ----------
    nconv : int
        successive autoconvolutions of the tophat. max=6.

    Thanks to Romke Bontekoe and Mathematica for providing the analytic expressions.

    """
    FWHM = [0.5, 0.5, 0.6339745962155612, 0.7223517244643762, 0.7971951696335494,
            0.8660920722545018, 0.9301994777857857]
    RANGE = [0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5]
    INTEGRAL = [1.0, 1.0, 4/3, 3/2, 192/115, 20/11, 11520/5887]


    def __init__( self, nconv=0 ) :
        """
        Constructor.

        Integral, fwhm and range are dependent on the number of convolutions.

        Parameters
        ----------
        nconv : int
            number of auto-convolutions

        """


        super( Tophat, self ).__init__( integral=self.INTEGRAL[nconv],
                        fwhm=2*self.FWHM[nconv], range=self.RANGE[nconv] )
        self.nconv = nconv

        if nconv == 0 :
            self.conv = Conv0()
        elif nconv == 1 :
            self.conv = Conv1()
        elif nconv == 2 :
            self.conv = Conv2()
        elif nconv == 3 :
            self.conv = Conv3()
        elif nconv == 4 :
            self.conv = Conv4()
        elif nconv == 5 :
            self.conv = Conv5()
        elif nconv == 6 :
            self.conv = Conv6()
        else :
            raise ValueError( "Cannot handle more than 6 convolutions." )

    def result( self, x ):
        return self.conv.result( x )

    def resultsq( self, xsq ):
        return self.result( xsq )                   #  the same

    def partial( self, x ):
        return self.conv.partial( x )

    def isBound( self ):
        return True

    def name( self ):
        return str( "Tophat %d convolved" % self.nconv )

class Conv0( object ) :
    def result( self, x ) :
        ax = numpy.abs( x )
        return numpy.where( ax < 0.5, 1.0, numpy.where( ax == 0.5, 0.5, 0.0 ) )

    def partial( self, x ):
        return numpy.zeros_like( x )

class Conv1( object ) :
    def result( self, x ) :
        ax = numpy.abs( x )
        return numpy.where( ax < 1.0, 1.0 - ax, 0.0 )

    def partial( self, x ):
        ax = numpy.abs( x )
        return numpy.where( ax < 1.0, - numpy.sign( x ), 0 )

class Conv2( object ) :

    def result( self, x ) :
        x = numpy.array( x, copy=False, ndmin=1 )
        ax = numpy.abs( x )
        q = numpy.where( ax < 1.5 )[0]
        res = numpy.zeros_like( x )
        if len( q ) == 0 :
            return res

        x1 = ax[q]
        x2 = x1 * x1

        b = numpy.where( x1 < 0.5 )
        res[q[b]] = ( 3 - 4 * x2[b] ) / 3,
        b = numpy.where( x1 >= 0.5 )
        res[q[b]] = ( 9 - 12 * x1[b] + 4 * x2[b] ) / 6

        return res

    def partial( self, x ):
        x = numpy.array( x, copy=False, ndmin=1 )
        ax = numpy.abs( x )
        q = numpy.where( ax < 1.5 )[0]
        res = numpy.zeros_like( x )
        if len( q ) == 0 :
            return res

        s0 = numpy.sign( x[q] )
        x1 = ax[q]

        b = numpy.where( x1 < 0.5 )
        res[q[b]] = s0[b] * ( -8 * x1[b] ) / 3
        b = numpy.where( x1 >= 0.5 )
        res[q[b]] = s0[b] * ( -12 + 8 * x1[b] ) / 6

        return res

class Conv3( object ) :
    def result( self, x ) :
        x = numpy.array( x, copy=False, ndmin=1 )
        ax = numpy.abs( x )
        q = numpy.where( ax < 2.0 )[0]
        res = numpy.zeros_like( x )
        if len( q ) == 0 :
            return res

        x1 = ax[q]
        x2 = x1 * x1
        x3 = x2 * x1

        b = numpy.where( x1 < 1.0 )
        res[q[b]] = ( 4 - 6 * x2[b] + 3 * x3[b] ) / 4
        b = numpy.where( x1 >= 1.0 )
        res[q[b]] = ( 8 - 12 * x1[b] + 6 * x2[b] - x3[b] ) / 4

        return res

    def partial( self, x ):
        x = numpy.array( x, copy=False, ndmin=1 )
        ax = numpy.abs( x )
        q = numpy.where( ax < 2.0 )[0]
        res = numpy.zeros_like( x )
        if len( q ) == 0 :
            return res

        s0 = numpy.sign( x[q] )
        x1 = ax[q]
        x2 = x1 * x1

        b = numpy.where( x1 < 1.0 )
        res[q[b]] = s0[b] * ( - 12 * x1[b] + 9 * x2[b] ) / 4
        b = numpy.where( x1 >= 1.0 )
        res[q[b]] = s0[b] * ( - 12 + 12 * x1[b] - 3 * x2[b] ) / 4

        return res

class Conv4( object ) :
    def result( self, x ) :
        x = numpy.array( x, copy=False, ndmin=1 )
        ax = numpy.abs( x )
        q = numpy.where( ax < 2.5 )[0]
        res = numpy.zeros_like( x )
        if len( q ) == 0 :
            return res

        x1 = ax[q]
        x2 = x1 * x1
        x3 = x2 * x1
        x4 = x3 * x1

        res[q] = ( 625 - 1000 * x1 + 600 * x2 - 160 * x3 + 16 * x4 ) / 230
        b = numpy.where( x1 < 1.5 )
        res[q[b]] = ( 55 + 20 * x1[b] - 120 * x2[b] + 80 * x3[b] - 16 * x4[b] ) * 2 / 115
        b = numpy.where( x1 < 0.5 )
        res[q[b]] = ( 115 - 120 * x2[b] +  48 * x4[b] ) / 115

        return res

    def partial( self, x ):
        x = numpy.array( x, copy=False, ndmin=1 )
        ax = numpy.abs( x )
        q = numpy.where( ax < 2.5 )[0]
        res = numpy.zeros_like( x )
        if len( q ) == 0 :
            return res

        s0 = numpy.sign( x[q] )
        x1 = ax[q]
        x2 = x1 * x1
        x3 = x2 * x1

        res[q] = s0 * ( - 1000 + 1200 * x1 - 480 * x2 + 64 * x3 ) / 230
        b = numpy.where( x1 < 1.5 )
        res[q[b]] = s0[b] * ( + 20  - 240 * x1[b] + 240 * x2[b] - 64 * x3[b] ) * 2 / 115
        b = numpy.where( x1 < 0.5 )
        res[q[b]] = s0[b] * ( - 240 * x1[b] +  192 * x3[b] ) / 115

        return res

class Conv5( object ) :
    def result( self, x ) :
        x = numpy.array( x, copy=False, ndmin=1 )
        ax = numpy.abs( x )
        q = numpy.where( ax < 3.0 )[0]
        res = numpy.zeros_like( x )
        if len( q ) == 0 :
            return res

        x1 = ax[q]
        x2 = x1 * x1
        x3 = x2 * x1
        x4 = x3 * x1
        x5 = x4 * x1

        res[q] = ( 243 - 405 * x1 + 270 * x2 - 90 * x3 + 15 * x4 - x5 ) / 66
        b = numpy.where( x1 < 2.0 )
        res[q[b]] = ( 51 + 75 * x1[b] - 210 * x2[b] + 150 * x3[b] - 45 * x4[b] + 5 * x5[b] ) / 66
        b = numpy.where( x1 < 1.0 )
        res[q[b]] = ( 33 - 30 * x2[b] + 15 * x4[b] - 5 * x5[b] ) / 33

        return res

    def partial( self, x ):
        x = numpy.array( x, copy=False, ndmin=1 )
        ax = numpy.abs( x )
        q = numpy.where( ax < 3.0 )[0]
        res = numpy.zeros_like( x )
        if len( q ) == 0 :
            return res

        s0 = numpy.sign( x[q] )
        x1 = ax[q]
        x2 = x1 * x1
        x3 = x2 * x1
        x4 = x3 * x1

        res[q] = s0 * ( - 405 + 540 * x1 - 270 * x2 + 60 * x3 - 5 * x4 ) / 66
        b = numpy.where( x1 < 2.0 )
        res[q[b]] = s0[b] * ( + 75 - 420 * x1[b] + 450 * x2[b] - 180 * x3[b] + 25 * x4[b] ) / 66
        b = numpy.where( x1 < 1.0 )
        res[q[b]] = s0[b] * ( - 60 * x1[b] + 60 * x3[b] - 25 * x4[b] ) / 33

        return res

class Conv6( object ) :
    def result( self, x ) :
        x = numpy.array( x, copy=False, ndmin=1 )
        ax = numpy.abs( x )
        q = numpy.where( ax < 3.5 )[0]
        res = numpy.zeros_like( x )
        if len( q ) == 0 :
            return res

        x1 = ax[q]
        x2 = x1 * x1
        x3 = x2 * x1
        x4 = x3 * x1
        x5 = x4 * x1
        x6 = x5 * x1

        b = numpy.where( x1 >= 2.5 )
        res[q[b]] = ( 117649 - 201684 * x1[b] + 144060 * x2[b] - 54880 * x3[b] + 11760 * x4[b] -
                    1344 * x5[b] + 64 * x6[b] ) / 23548
        b = numpy.where( numpy.logical_and( x1 < 2.5, x1 >= 1.5 ) )
        res[q[b]] = ( 4137 + 30408 * x1[b] - 59220 * x2[b] + 42560 * x3[b] - 15120 * x4[b] +
                    2688 * x5[b] - 192 * x6[b] ) / 11774
        b = numpy.where( numpy.logical_and( x1 < 1.5, x1 >= 0.5 ) )
        res[q[b]] = ( 23583 - 420 * x1[b] - 16380 * x2[b] - 5600 * x3[b] + 15120 * x4[b] -
                    6720 * x5[b] + 960 * x6[b] ) / 23548
        b = numpy.where( x1 < 0.5 )
        res[q[b]] = ( 5887 - 4620 * x2[b] + 1680 * x4[b] - 320 * x6[b] ) / 5887

        return res

    def partial( self, x ):
        x = numpy.array( x, copy=False, ndmin=1 )
        ax = numpy.abs( x )
        q = numpy.where( ax < 3.5 )[0]
        res = numpy.zeros_like( x )
        if len( q ) == 0 :
            return res

        s0 = numpy.sign( x[q] )
        x1 = ax[q]
        x2 = x1 * x1
        x3 = x2 * x1
        x4 = x3 * x1
        x5 = x4 * x1

        b = numpy.where( x1 >= 2.5 )
        res[q[b]] = s0[b] * ( - 201684 + 2*144060 * x1[b] - 3*54880 * x2[b] + 4*11760 * x3[b] -
                    5*1344 * x4[b] + 6*64 * x5[b] ) / 23548
        b = numpy.where( numpy.logical_and( x1 < 2.5, x1 >= 1.5 ) )
        res[q[b]] = s0[b] * ( + 30408 - 2*59220 * x1[b] + 3*42560 * x2[b] - 4*15120 * x3[b] +
                    5*2688 * x4[b] - 6*192 * x5[b] ) / 11774
        b = numpy.where( numpy.logical_and( x1 < 1.5, x1 >= 0.5 ) )
        res[q[b]] = s0[b] * ( - 420 - 2*16380 * x1[b] - 3*5600 * x2[b] + 4*15120 * x3[b] -
                    5*6720 * x4[b] + 6*960 * x5[b] ) / 23548
        b = numpy.where( x1 < 0.5 )
        res[q[b]] = s0[b] * ( - 2*4620 * x1[b] + 4*1680 * x3[b] - 6*320 * x5[b] ) / 5887

        return res








import numpy as numpy
import math as math
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from numpy.testing import assert_array_almost_equal as assertAAE
from . import Tools
from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

# This file is part of the BayesicFitting package.
#
# BayesicFitting is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or ( at your option ) any later version.
#
# BayesicFitting is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#
#   2017 - 2022 Do Kester

def plotFit( x, data=None, yfit=None, model=None, fitter=None, show=True,
             residuals=False, xlabel=None, ylabel=None, title=None, figsize=[7,5], 
             xlim=None, ylim=None, filename=None, transparent=False ) :
    """
    Plot the data of a fit.

    Parameters
    ----------
    x : array_like
        xdata of the problem
    data : array_like
        ydata of the problem
    yfit : array_like
        fit of the data to the model
    model : Model
        the model the data are fitted to at x
    fitter : BaseFitter
        the fitter being used
        If set it displays a confidence region for the fit
    show : bool
        display the plot
    residuals : bool
        plot the residuals in a separate panel
    xlabel : None or str
        use as xlabel
    ylabel : None or str
        use as ylabel
    title  : None or str
        use as title
    xlim : None or list of 2 floats
        limits on x-axis
    ylim : None or list of 2 floats
        limits on y-axis
    figsize : list of 2 floats
        size of the figure
    filename  : None or str
        name of png file; otherwise show
    transparent : bool
        make the png file transparent
    """

    if xlabel is None : xlabel = "xdata"
    if ylabel is None : ylabel = "ydata"

    plt.figure( "Fitter Results", figsize=figsize )

    minx = numpy.min( x )
    maxx = numpy.max( x )
    ax0 = plt

    if yfit is None and model is not None :
        yfit = model( x )

    if residuals :
        plt.subplots_adjust( hspace=0.001 )
        gs = gridspec.GridSpec( 2, 1, height_ratios=[4, 1])

        ax1 = plt.subplot( gs[1] )
        res = data - yfit
        nd = int( math.log10( len( data ) ) )
        mrksz = ( 5 - nd ) if nd < 4 else 1
        ax1.plot( x, res, 'k.', markersize=mrksz )
        ax1.margins( 0.05, 0.05 )
        xt = plt.ylim()
        xtk = Tools.nicenumber( ( xt[1] - xt[0] ) / 4 )

        plt.yticks( [-xtk, 0.0, xtk] )
        plt.ylabel( "residual" )
        plt.xlabel( xlabel )

        ax0 = plt.subplot( gs[0] )
        xticklabels = ax0.get_xticklabels()
        plt.setp( xticklabels, visible=False )

    if data is not None :
        nd = int( math.log10( len( data ) ) )
        mrksz = ( 5 - nd ) if nd < 4 else 1
        ax0.plot( x, data, 'k.', markersize=mrksz )

    if model is not None :
        xx = numpy.linspace( minx, maxx, 10 * len( x ) )
        yy = model( xx )
        if fitter is not None :
            err = fitter.monteCarloError( xx )
            ax0.plot( xx, yy - err, 'g-' )
            ax0.plot( xx, yy + err, 'g-' )
        ax0.plot( xx, yy, 'r-' )
    elif yfit is not None :
        ax0.plot( x, yfit, 'r-' )

    ax0.margins( 0.05, 0.05 )
    plt.ylabel( ylabel )
    if not residuals :
        plt.xlabel( xlabel )

    if xlim is not None :
        plt.xlim( xlim[0], xlim[1] )
    if ylim is not None :
        plt.ylim( ylim[0], ylim[1] )

    if title is not None :
        plt.title( title )

    if filename is None :
        if show : plt.show()
    else :
        plt.savefig( filename, transparent=transparent )




def plotSampleList( sl, xdata, ydata, errors=None, npt=10000,
        residuals=False, xlabel=None, ylabel=None, title=None, figsize=[7,5], 
        xlim=None, ylim=None, filename=None, transparent=False, show=True ) :
    """
    Plot the posterior as npt points from the SampleList.

    Parameters
    ==========
    sl : SampleList
        the samplelist containing samples from the posterior
    xdata : arraylike
        the xdata values; plotted for comparison
    ydata : arraylike
        the ydata values; plotted for comparison
    errors : None of arraylike
        (No) errors on the ydata are displayed
    npt : int
        number of points from the sample (10000)
    residuals : bool
        plot the residuals in a lower panel (False)
    xlabel : None or str
        use as xlabel
    ylabel : None or str
        use as ylabel
    title  : None or str
        use as title
    xlim : None or list of 2 floats
        limits on x-axis
    ylim : None or list of 2 floats
        limits on y-axis
    figsize : list of 2 floats
        size of the figure
    filename  : None or str
        name of png file; otherwise show
    transparent : bool
        make the png file transparent

    """
    if xlabel is None : xlabel = "xdata"
    if ylabel is None : ylabel = "ydata"

    plt.figure( figsize=figsize )
    ax0 = plt

    km = sl.medianIndex
    smp = sl[km]

    if residuals :
        plt.subplots_adjust( hspace=0.001 )
        gs = gridspec.GridSpec( 2, 1, height_ratios=[4,1] )

        ax1 = plt.subplot( gs[1] )
        # plot zero line
        ax1.plot( [min( xdata ),max( xdata )], [0,0], 'k-' )

        res = ydata - sl.average( xdata )
        nd = int( math.log10( len( ydata ) ) )
        mrksz = ( 5 - nd ) if nd < 4 else 1
        ax1.plot( xdata, res, 'k.', markersize=mrksz )
        ax1.margins( 0.05, 0.05 )
        yt = plt.ylim()
        ymx = max( abs( yt[0] ), abs( yt[1] ) )
        ytk = Tools.nicenumber( 0.7 * ymx )

#        print( "Plotter  ", fmt( yt ), fmt( ymx ), fmt( ytk ) )
 
        plt.yticks( [-ytk, 0.0, ytk] )
        plt.ylim( -ymx, ymx )
        plt.ylabel( "residuals" )
        plt.xlabel( xlabel )

        ax0 = plt.subplot( gs[0] )
        xticklabels = ax0.get_xticklabels()
        plt.setp( xticklabels, visible=False )


    if xlim is None : 
        xlo = numpy.min( xdata )
        xhi = numpy.max( xdata )
    else :
        xlo, xhi = tuple( xlim )
    xrng = xhi - xlo

    # plot npt samples from posterior
    ntot = 0
    swgt = 0.0

    for s in sl :
        swgt += s.weight * npt
        ns = int( swgt )
        if ns == 0 : continue
    
        #print( s.weight, swgt, ns )
        xs = numpy.random.rand( ns ) * xrng + xlo

        ax0.plot( xs, s.model.result( xs, s.parameters ), 'r.', markersize=2 )
        swgt -= ns
        ntot += ns

    # plot data
    if errors is None :
        ax0.plot( xdata, ydata, 'k.' )
    else :
        ax0.errorbar( xdata, ydata, yerr=errors, fmt='k.' )
    
    # plot average in green
    xs = numpy.linspace( xlo, xhi, 201 )
    yfit = sl.average( xs )
    ax0.plot( xs, yfit, 'g-' )

    plt.ylabel( ylabel )
    if not residuals :
        plt.xlabel( xlabel )

    if ylim is None :
        ylo = min( min( ydata ), min( yfit ) )
        yhi = max( max( ydata ), max( yfit ) )
        ylo -= 0.05 * ( yhi - ylo )
        yhi += 0.05 * ( yhi - ylo )

    if xlim is None :
        xlo -= 0.05 * xrng
        xhi += 0.05 * xrng

    plt.xlim( xlo, xhi )
    plt.ylim( ylo, yhi )

    if title is not None :
        plt.title( title )

    if filename is None :
        if show : plt.show()
    else :
        plt.savefig( filename, transparent=transparent )





# run with : python3 -m unittest TestOrthonormalBasis

import unittest
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt
from numpy.testing import assert_array_almost_equal as assertAAE

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestOrthonormalBasis( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author       Do Kester

    """
    def __init__( self, name ):
        super( TestOrthonormalBasis, self ).__init__( name )


    def testbasis( self ) :
        """
        Numeric values from
        http://www.ecs.umass.edu/ece/ece313/Online_help/gram.pdf

        """

        print( "===== orthonormal basis ================" )

        onb = OrthonormalBasis()

        a1 = numpy.array( [1,0,2,1], dtype=float )
        u1 = onb.normalise( a1 )
        print( "vector  ", fmt( a1 ) )
        print( "ortnor  ", fmt( u1 ) )
        print( "basis  ", fmt( onb.basis, indent=8 ) )
        print( "" )
        assertAAE( u1, a1 / math.sqrt( 6 ) )

        a2 = numpy.array( [2,2,3,1], dtype=float )
        u2 = onb.normalise( a2 )
        print( "vector  ", fmt( a2 ) )
        print( "ortnor  ", fmt( u2 ) )
        print( "basis  ", fmt( onb.basis, indent=8 ) )
        print( "" )
        assertAAE( u2, numpy.array( [1,4,0,-1] ) / ( 3 * math.sqrt( 2 ) ) )

        a3 = numpy.array( [1,1,0,1], dtype=float )
        u3 = onb.normalise( a3 )
        print( "vector  ", fmt( a3 ) )
        print( "ortnor  ", fmt( u3 ) )
        print( "basis  ", fmt( onb.basis, indent=8 ) )
        print( "" )
        assertAAE( u3, numpy.array( [4,1,-6,8] )/ math.sqrt( 117 ) )

        a4 = numpy.array( [2,1,0,1], dtype=float )
        u4 = onb.normalise( a4 )
        print( "vector  ", fmt( a4 ) )
        print( "ortnor  ", fmt( u4 ) )
        print( "basis  ", fmt( onb.basis, indent=8 ) )
        print( "" )
#        assertAAE( u4, numpy.array( [4,1,-6,8] )/ math.sqrt( 117 ) )

        b = onb.basis
        assertAAE( numpy.inner( b[0,:], b[1,:] ), 0.0 )
        assertAAE( numpy.inner( b[0,:], b[2,:] ), 0.0 )
        assertAAE( numpy.inner( b[0,:], b[3,:] ), 0.0 )
        assertAAE( numpy.inner( b[1,:], b[2,:] ), 0.0 )
        assertAAE( numpy.inner( b[1,:], b[3,:] ), 0.0 )
        assertAAE( numpy.inner( b[2,:], b[3,:] ), 0.0 )

        a5 = numpy.array( [1,1,0,2], dtype=float )
        u5 = onb.normalise( a5 )
        print( "vector  ", fmt( a5 ) )
        print( "ortnor  ", fmt( u5 ) )
        print( "basis  ", fmt( onb.basis, indent=8 ) )
        print( "" )
        assertAAE( u5, numpy.array( [1,1,0,2] )/ math.sqrt( 6 ) )

        a6 = numpy.random.rand( 20 ) - 0.5
        u6 = onb.normalise( a6, reset=True )
        print( "vector  ", fmt( a6 ) )
        print( "ortnor  ", fmt( u6 ) )
        print( "basis  ", fmt( onb.basis, indent=8, max=None ) )
        print( "" )

        a7 = numpy.random.rand( 20 ) - 0.5
        u7 = onb.normalise( a7, reset=False )
        print( "vector  ", fmt( a7 ) )
        print( "ortnor  ", fmt( u7 ) )
        print( "basis  ", fmt( onb.basis, indent=8, max=None ) )
        print( "" )

        b = onb.basis
        assertAAE( numpy.inner( b[0,:], b[1,:] ), 0.0 )



    def suite( cls ):
        return unittest.TestCase.suite( TestOrthonormalBasis.__class__ )


if __name__ == '__main__':
    unittest.main( )





import numpy as numpy
from astropy import units
import math
from . import Tools
from .Tools import setAttribute as setatt
from .Formatter import formatter as fmt


__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"


#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2021        Do Kester

##############################################################################
##  This utilities contain a number of classes that can be combined to a    ##
##  Neural Net of arbitrary complexity.                                     ##
##                                                                          ##
##  A Neural Net Model consists of at least one hidden layer of neurons.    ##
##  The inputs are linearly scaled on the first layer.                      ##
##  Optionally a bias can be added to the neurons.                          ##
##  A transfer function is applied to the neurons.                          ##
##  The layer acts as input for the next layer where the same procedure is  ##
##  followed. Until finally it is projected on the output layer.            ##
##                                                                          ##
##############################################################################



class BaseNeuralNetworkUtilities( object ) :


    def pipe( self, dfdr, drdp ) :
        dfdp = numpy.zeros_like( drdp )

        ( ndout, ndata, nip ) = drdp.shape
        for k in range( nip ) :
            for i in range( ndata ) :
                dfdp[:,i,k] = numpy.inner( dfdr[:,i,:], drdp[:,i,k] )

        if ndout == 1 :
            return dfdp[0,:,:]

        return dfdp


class Connect( BaseNeuralNetworkUtilities ):
    """
    Connects one layer to the next.

        0       0       0       0       0       0   K inputs
        |\     /|\     /|\     /|\     /|\     /|

          all inputs connect to all outputs         K*N parameters

           \|/     \|/     \|/     \|/     \|/      N offsets (see ConnectWithBias)
            0       0       0       0       0       N outputs


    Connects layer1 to layer2 via a matrix of [K*N]

    The parameters are initialized at {0.0}.


    Attributes
    ----------
        ndim : int
            number of input parameters
        ndout : int
            number of output parameters
        nraster : int
            ndim * ndout

    """

    def __init__( self, ndim=1, ndout=1 ):

        """
        (Part of) Neural Net.

        Parameters
        ----------
        ndim : int
            number of input columns
        ndout : int
            number of output values, per input row. See also ncat

        """
        self.ndim = ndim
        self.ndout = ndout
        self.nraster = ndim * ndout

        super( ).__init__( )

    def result( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        par = numpy.reshape( params, ( self.ndout, self.ndim ) )
        xd  = numpy.reshape( xdata, ( -1, self.ndim ) )
        return numpy.inner( xd, par )

    def partial( self, xdata, params ) :

        shp = ( self.ndout, Tools.length( xdata ), len( params ) )
        partial = numpy.zeros( shp, dtype=float )

        i2o = numpy.arange( self.ndim, dtype=int )

        for n in range( self.ndout ) :
            partial[n,:,i2o] = xdata.T.copy()
            i2o += self.ndim

        return partial
#        return partial if self.ndout > 1 else partial[0,:,:]
  

    def derivative( self, xdata, params ) :

        nx = Tools.length( xdata )
        dfdx = numpy.zeros( ( self.ndout, nx, self.ndim ), dtype=float )

        par = params.reshape( self.ndout, self.ndim )

        for n in range( self.ndout ) :
            for k in range( self.ndim ) :
               dfdx[n,:,k] = par[n,k]

        return dfdx

class ConnectWithBias( Connect ) :
    """
    Connect as with the previous Connect class and add a bias to layer2
    Using the next N parameters
    """

    def __init__( self, ndim=1, ndout=1 ):
        super( ).__init__( ndim=ndim, ndout=ndout )


    def result( self, xdata, params ):
        return super().result( xdata, params[:self.nraster] ) + params[self.nraster:]

    def partial( self, xdata, params ) :
        pt1 = super().partial( xdata, params[:self.nraster] )
        shp = list( pt1.shape )
        shp[2] += self.ndout
        part = numpy.zeros( shp, dtype=float )
        for k in range( self.ndout ) :
            pt2 = numpy.zeros( ( Tools.length( xdata ), self.ndout ), dtype=float )
            pt2[:,k] = 1
            part[k] = numpy.append( pt1[k], pt2, axis=1 )    

        return part

    def derivative( self, xdata, params ) :
        return super().derivative( xdata, params[:self.nraster] )


######################################################
##                                                  ##
## Hereafter follow a number of transfer functions. ##
##                                                  ##
## Partials are never needed as they dont have      ##
## parameters                                       ##
##                                                  ##
######################################################


class Identity( BaseNeuralNetworkUtilities ) :
    """ 
    Transfer
        f(x) = x 
    """

    def result( self, xdata, params ) :
        return xdata

    def derivative( self, xdata, params ) :
        return numpy.ones_like( xdata )

    def pipe( self, dfdr, drdp ) :
        # dfdr is all 1's
        return drdp


class Arctan( BaseNeuralNetworkUtilities ) :
    """ 
    Transfer
        f(x) = arctan( x ) 
    """

    def result( self, xdata, params ) :
        return numpy.arctan( xdata )

    def derivative( self, xdata, params ) :
        return 1 / ( 1 + xdata * xdata )

    def pipe( self, dfdr, drdp ) :
        dfdp = numpy.zeros_like( drdp )
        dfdrt = dfdr.T
        for k in range( drdp.shape[-1] ) :
            dfdp[:,:,k] = dfdrt * drdp[:,:,k]

        return dfdp


class Logistic( BaseNeuralNetworkUtilities ) :
    """
    To covert outputs to categorical probabilities.
    For 1 choice (category) only: True or False

    Transfer
        f(x) = 1.0 / ( 1.0 + exp( -x ) )

    """
    def result( self, xdata, params ) :
        return 1.0 / ( 1.0 + numpy.exp( -xdata ) )

    def derivative( self, xdata, params ) :
        e = numpy.exp( xdata )
        return e / numpy.square( 1 + e )

    def pipe( self, dfdr, drdp ) :
        dfdp = numpy.zeros_like( drdp )
        dfdrt = dfdr.T
        for k in range( drdp.shape[-1] ) :
            dfdp[:,:,k] = dfdrt * drdp[:,:,k]

        return dfdp

class Heaviside( BaseNeuralNetworkUtilities ) :
    """ 
    Transfer 
        stepfunction at 0 
    """

    def result( self, xdata, params ) :
        return numpy.where( xdata >=0, 1.0, 0.0 )

    def derivative( self, xdata, params ) :
        return numpy.zeros_like( xdata )

    def pipe( self, dfdr, drdp ) :
        return numpy.zeros_like( drdp )

class Softmax( BaseNeuralNetworkUtilities ) :
    """
    To covert outputs to categorical probabilities.
    For more choices (categories): red, green, blue, etc.

    The probabilities are normalized: SUM( f ) = 1

    Transfer 
        f( x_n ) = exp( x_n ) / SUM_i( exp( x_i ) )
    """ 
    def __init__( self, ndout=1 ) :
        self.ndout = ndout

    def result( self, xdata, params ) :
        res = numpy.exp( xdata )
        sum = numpy.sum( res, axis=1 )
        return ( res.T / sum ).T

    def derivative( self, xdata, params ) :
        f = numpy.exp( xdata )
        ndata = xdata.shape[0]

        norm = numpy.sum( f, axis=1 )
        n2 = norm * norm

        shp = ( self.ndout, ndata, self.ndout )
        dfdx = numpy.zeros( shp, dtype=float )

        for n in range( self.ndout ) :
            for k in range( self.ndout ) :
                dfdx[k,:,n] = - ( f[:,k] * f[:,n] / n2 )
            dfdx[n,:,n] = ( ( norm - f[:,n]  ) * f[:,n] / n2 ).T

        return dfdx




